from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.checks import Error
from django.db import models
from django.db.utils import IntegrityError, OperationalError
from django.utils import timezone

from .exceptions import OAuthUserAlreadyExistsError

if TYPE_CHECKING:
    from .providers import OAuthToken, OAuthUser


# django check for deploy that ensures all provider keys in db are also in settings?


class OAuthConnection(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="oauth_connections",
    )

    # The key used to refer to this provider type (in settings)
    provider_key = models.CharField(max_length=100, db_index=True)

    # The unique ID of the user on the provider's system
    provider_user_id = models.CharField(max_length=100, db_index=True)

    # Token data
    access_token = models.CharField(max_length=100, blank=True)
    refresh_token = models.CharField(max_length=100, blank=True)
    access_token_expires_at = models.DateTimeField(blank=True, null=True)
    refresh_token_expires_at = models.DateTimeField(blank=True, null=True)

    class Meta:
        unique_together = ("provider_key", "provider_user_id")
        ordering = ("provider_key",)

    def __str__(self):
        return f"{self.provider_key}[{self.user}:{self.provider_user_id}]"

    def refresh_access_token(self) -> None:
        from .providers import OAuthToken, get_oauth_provider_instance

        provider_instance = get_oauth_provider_instance(provider_key=self.provider_key)
        oauth_token = OAuthToken(
            access_token=self.access_token,
            refresh_token=self.refresh_token,
            access_token_expires_at=self.access_token_expires_at,
            refresh_token_expires_at=self.refresh_token_expires_at,
        )
        refreshed_oauth_token = provider_instance.refresh_oauth_token(
            oauth_token=oauth_token
        )
        self.set_token_fields(refreshed_oauth_token)

    def set_token_fields(self, oauth_token: "OAuthToken"):
        self.access_token = oauth_token.access_token
        self.refresh_token = oauth_token.refresh_token
        self.access_token_expires_at = oauth_token.access_token_expires_at
        self.refresh_token_expires_at = oauth_token.refresh_token_expires_at

    def set_user_fields(self, oauth_user: "OAuthUser"):
        self.provider_user_id = oauth_user.id

    def can_be_disconnected(self) -> bool:
        return (
            self.user.has_usable_password() or self.user.oauth_connections.count() > 1
        )

    def access_token_expired(self) -> bool:
        return (
            self.access_token_expires_at is not None
            and self.access_token_expires_at < timezone.now()
        )

    def refresh_token_expired(self) -> bool:
        return (
            self.refresh_token_expires_at is not None
            and self.refresh_token_expires_at < timezone.now()
        )

    @classmethod
    def get_or_createuser(
        cls, *, provider_key: str, oauth_token: "OAuthToken", oauth_user: "OAuthUser"
    ) -> "OAuthConnection":
        try:
            connection = cls.objects.get(
                provider_key=provider_key,
                provider_user_id=oauth_user.id,
            )
            connection.set_token_fields(oauth_token)
            connection.save()
            return connection
        except cls.DoesNotExist:
            # If email needs to be unique, then we expect
            # that to be taken care of on the user model itself
            try:
                user = get_user_model().objects.create_user(
                    username=oauth_user.username,
                    email=oauth_user.email,
                )
            except IntegrityError:
                raise OAuthUserAlreadyExistsError()

            return cls.connect(
                user=user,
                provider_key=provider_key,
                oauth_token=oauth_token,
                oauth_user=oauth_user,
            )

    @classmethod
    def connect(
        cls,
        *,
        user: settings.AUTH_USER_MODEL,
        provider_key: str,
        oauth_token: "OAuthToken",
        oauth_user: "OAuthUser",
    ) -> "OAuthConnection":
        """
        Connect will either create a new connection or update an existing connection
        """
        connection, _ = cls.objects.get_or_create(
            user=user,
            provider_key=provider_key,
            provider_user_id=oauth_user.id,
        )
        connection.set_user_fields(oauth_user)
        connection.set_token_fields(oauth_token)
        connection.save()
        return connection

    @classmethod
    def check(cls, **kwargs):
        """
        A system check for ensuring that provider_keys in the database are also present in settings.

        Note that the --database flag is required for this to work:
          python manage.py check --database default
        """
        errors = super().check(**kwargs)

        databases = kwargs.get("databases", None)
        if not databases:
            return errors

        from .providers import get_provider_keys

        for database in databases:
            try:
                keys_in_db = set(
                    cls.objects.using(database)
                    .values_list("provider_key", flat=True)
                    .distinct()
                )
            except OperationalError:
                # Check runs on manage.py migrate, and the table may not exist yet
                # or it may not be installed on the particular database intentionally
                continue

            keys_in_settings = set(get_provider_keys())

            if keys_in_db - keys_in_settings:
                errors.append(
                    Error(
                        "The following OAuth providers are in the database but not in the settings: {}".format(
                            ", ".join(keys_in_db - keys_in_settings)
                        ),
                        id="oauth_login.E001",
                    )
                )

        return errors



from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import redirect, render
from django.views import View

from .exceptions import (
    OAuthCannotDisconnectError,
    OAuthStateMismatchError,
    OAuthUserAlreadyExistsError,
)
from .providers import get_oauth_provider_instance


class OAuthLoginView(View):
    def post(self, request, provider):
        if request.user.is_authenticated:
            return redirect("/")

        provider_instance = get_oauth_provider_instance(provider_key=provider)
        return provider_instance.handle_login_request(request=request)


class OAuthCallbackView(View):
    """
    The callback view is used for signup, login, and connect.
    """

    def get(self, request, provider):
        provider_instance = get_oauth_provider_instance(provider_key=provider)
        try:
            return provider_instance.handle_callback_request(request=request)
        except OAuthUserAlreadyExistsError:
            return render(
                request,
                "oauth_login/error.html",
                {
                    "oauth_error": "A user already exists with this email address. Please log in first and then connect this OAuth provider to the existing account."
                },
                status=400,
            )
        except OAuthStateMismatchError:
            return render(
                request,
                "oauth_login/error.html",
                {"oauth_error": "The state parameter did not match. Please try again."},
                status=400,
            )


class OAuthConnectView(LoginRequiredMixin, View):
    def post(self, request, provider):
        provider_instance = get_oauth_provider_instance(provider_key=provider)
        return provider_instance.handle_connect_request(request=request)


class OAuthDisconnectView(LoginRequiredMixin, View):
    def post(self, request, provider):
        provider_instance = get_oauth_provider_instance(provider_key=provider)
        try:
            return provider_instance.handle_disconnect_request(request=request)
        except OAuthCannotDisconnectError:
            return render(
                request,
                "oauth_login/error.html",
                {
                    "oauth_error": "This connection can't be removed. You must have a usable password or at least one active connection."
                },
                status=400,
            )



import datetime
import secrets
from typing import List
from urllib.parse import urlencode

from django.conf import settings
from django.contrib.auth import login
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.module_loading import import_string

from .exceptions import OAuthCannotDisconnectError, OAuthStateMismatchError
from .models import OAuthConnection

SESSION_STATE_KEY = "oauth_login_state"
SESSION_NEXT_KEY = "oauth_login_next"


class OAuthToken:
    def __init__(
        self,
        *,
        access_token: str,
        refresh_token: str = "",
        access_token_expires_at: datetime.datetime = None,
        refresh_token_expires_at: datetime.datetime = None,
    ):
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.access_token_expires_at = access_token_expires_at
        self.refresh_token_expires_at = refresh_token_expires_at


class OAuthUser:
    def __init__(self, *, id: str, email: str, username: str = ""):
        self.id = id
        self.username = username
        self.email = email


class OAuthProvider:
    authorization_url = ""

    def __init__(
        self,
        *,
        # Provided automatically
        provider_key: str,
        # Required as kwargs in OAUTH_LOGIN_PROVIDERS setting
        client_id: str,
        client_secret: str,
        # Not necessarily required, but commonly used
        scope: str = "",
    ):
        self.provider_key = provider_key
        self.client_id = client_id
        self.client_secret = client_secret
        self.scope = scope

    def get_authorization_url_params(self, *, request: HttpRequest) -> dict:
        return {
            "redirect_uri": self.get_callback_url(request=request),
            "client_id": self.get_client_id(),
            "scope": self.get_scope(),
            "state": self.generate_state(),
            "response_type": "code",
        }

    def refresh_oauth_token(self, *, oauth_token: OAuthToken) -> OAuthToken:
        raise NotImplementedError()

    def get_oauth_token(self, *, code: str, request: HttpRequest) -> OAuthToken:
        raise NotImplementedError()

    def get_oauth_user(self, *, oauth_token: OAuthToken) -> OAuthUser:
        raise NotImplementedError()

    def get_authorization_url(self, *, request: HttpRequest) -> str:
        return self.authorization_url

    def get_client_id(self) -> str:
        return self.client_id

    def get_client_secret(self) -> str:
        return self.client_secret

    def get_scope(self) -> str:
        return self.scope

    def get_callback_url(self, *, request: HttpRequest) -> str:
        url = reverse("oauth_login:callback", kwargs={"provider": self.provider_key})
        return request.build_absolute_uri(url)

    def generate_state(self) -> str:
        return get_random_string(length=32)

    def check_request_state(self, *, request: HttpRequest) -> None:
        state = request.GET["state"]
        expected_state = request.session.pop(SESSION_STATE_KEY)
        if not secrets.compare_digest(state, expected_state):
            raise OAuthStateMismatchError()

    def handle_login_request(self, *, request: HttpRequest) -> HttpResponse:
        authorization_url = self.get_authorization_url(request=request)
        authorization_params = self.get_authorization_url_params(request=request)

        if "state" in authorization_params:
            # Store the state in the session so we can check on callback
            request.session[SESSION_STATE_KEY] = authorization_params["state"]

        if "next" in request.POST:
            # Store in session so we can get it on the callback request
            request.session[SESSION_NEXT_KEY] = request.POST["next"]

        # Sort authorization params for consistency
        sorted_authorization_params = sorted(authorization_params.items())
        redirect_url = authorization_url + "?" + urlencode(sorted_authorization_params)
        return HttpResponseRedirect(redirect_url)

    def handle_connect_request(self, *, request: HttpRequest) -> HttpResponse:
        return self.handle_login_request(request=request)

    def handle_disconnect_request(self, *, request: HttpRequest) -> HttpResponse:
        provider_user_id = request.POST["provider_user_id"]
        connection = OAuthConnection.objects.get(
            provider_key=self.provider_key, provider_user_id=provider_user_id
        )
        if (
            request.user.has_usable_password()
            or request.user.oauth_connections.count() > 1
        ):
            connection.delete()
        else:
            raise OAuthCannotDisconnectError(
                "Cannot remove last OAuth connection without a usable password"
            )

        redirect_url = self.get_disconnect_redirect_url(request=request)
        return HttpResponseRedirect(redirect_url)

    def handle_callback_request(self, *, request: HttpRequest) -> HttpResponse:
        self.check_request_state(request=request)

        oauth_token = self.get_oauth_token(code=request.GET["code"], request=request)
        oauth_user = self.get_oauth_user(oauth_token=oauth_token)

        if request.user.is_authenticated:
            connection = OAuthConnection.connect(
                user=request.user,
                provider_key=self.provider_key,
                oauth_token=oauth_token,
                oauth_user=oauth_user,
            )
            user = connection.user
        else:
            connection = OAuthConnection.get_or_createuser(
                provider_key=self.provider_key,
                oauth_token=oauth_token,
                oauth_user=oauth_user,
            )

            user = connection.user

            # Log them in
            login(request, user)

        redirect_url = self.get_login_redirect_url(request=request)
        return HttpResponseRedirect(redirect_url)

    def get_login_redirect_url(self, *, request: HttpRequest) -> str:
        return request.session.pop(SESSION_NEXT_KEY, settings.LOGIN_REDIRECT_URL)

    def get_disconnect_redirect_url(self, *, request: HttpRequest) -> str:
        return request.POST.get("next", "/")


def get_oauth_provider_instance(*, provider_key: str) -> OAuthProvider:
    OAUTH_LOGIN_PROVIDERS = getattr(settings, "OAUTH_LOGIN_PROVIDERS", {})
    provider_class_path = OAUTH_LOGIN_PROVIDERS[provider_key]["class"]
    provider_class = import_string(provider_class_path)
    provider_kwargs = OAUTH_LOGIN_PROVIDERS[provider_key].get("kwargs", {})
    return provider_class(provider_key=provider_key, **provider_kwargs)


def get_provider_keys() -> List[str]:
    return list(getattr(settings, "OAUTH_LOGIN_PROVIDERS", {}).keys())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oauth_login', 'oauth_login.migrations']

package_data = \
{'': ['*'], 'oauth_login': ['templates/oauth_login/*']}

setup_kwargs = {
    'name': 'django-oauth-login',
    'version': '0.1.0',
    'description': '',
    'long_description': '# django-oauth-login\n\n**Add OAuth login support to your Django project.**\n\nThis library is intentionally minimal.\nIt has no dependencies and a single database model.\nIf you simply want users to log in with GitHub, Google, Twitter, etc. (and maybe use that access token for API calls),\nthen this is the library for you.\n\nThere are three OAuth flows that it makes possible:\n\n1. Signup via OAuth (new user, new OAuth connection)\n2. Login via OAuth (existing user, existing OAuth connection)\n3. Connect/disconnect OAuth accounts to a user (existing user, new OAuth connection)\n\n\n## Usage\n\nInstall the package from PyPi:\n\n```sh\npip install django-oauth-login\n```\n\nAdd `oauth_login` to your `INSTALLED_APPS` in `settings.py`:\n\n```python\nINSTALLED_APPS = [\n    ...\n    "oauth_login",\n]\n```\n\nIn your `urls.py`, include `oauth_login.urls`:\n\n```python\nurlpatterns = [\n    path("oauth/", include("oauth_login.urls")),\n    ...\n]\n```\n\nCreate a new OAuth provider ([or copy one from our examples](provider_examples)):\n\n```python\n# yourapp/oauth.py\nimport requests\n\nfrom oauth_login.providers import OAuthProvider, OAuthToken, OAuthUser\n\n\nclass ExampleOAuthProvider(OAuthProvider):\n    authorization_url = "https://example.com/login/oauth/authorize"\n\n    def get_oauth_token(self, *, code, request):\n        response = requests.post(\n            "https://example.com/login/oauth/token",\n            headers={\n                "Accept": "application/json",\n            },\n            data={\n                "client_id": self.get_client_id(),\n                "client_secret": self.get_client_secret(),\n                "code": code,\n            },\n        )\n        response.raise_for_status()\n        data = response.json()\n        return OAuthToken(\n            access_token=data["access_token"],\n        )\n\n    def get_oauth_user(self, *, oauth_token):\n        response = requests.get(\n            "https://example.com/api/user",\n            headers={\n                "Accept": "application/json",\n                "Authorization": f"token {oauth_token.access_token}",\n            },\n        )\n        response.raise_for_status()\n        data = response.json()\n        return OAuthUser(\n            id=data["id"],\n            username=data["username"],\n            email=data["email"],\n        )\n```\n\nCreate your OAuth app/consumer on the provider\'s site (GitHub, Google, etc.).\nWhen setting it up, you\'ll likely need to give it a callback URL.\nIn development this can be `http://localhost:8000/oauth/github/callback/` (if you name it `"github"` like in the example below).\nAt the end you should get some sort of "client id" and "client secret" which you can then use in your `settings.py`:\n\n```python\nOAUTH_LOGIN_PROVIDERS = {\n    "github": {\n        "class": "yourapp.oauth.GitHubOAuthProvider",\n        "kwargs": {\n            "client_id": environ["GITHUB_CLIENT_ID"],\n            "client_secret": environ["GITHUB_CLIENT_SECRET"],\n            # "scope" is optional, defaults to ""\n\n            # You can add other fields if you have additional kwargs in your class __init__\n            # def __init__(self, *args, custom_arg="default", **kwargs):\n            #     self.custom_arg = custom_arg\n            #     super().__init__(*args, **kwargs)\n        },\n    },\n}\n```\n\nThen add a login button (which is a form using POST rather than a basic link, for security purposes):\n\n```html\n<h1>Login</h1>\n<form action="{% url \'oauth_login:login\' \'github\' %}" method="post">\n    {% csrf_token %}\n    <button type="submit">Login with GitHub</button>\n</form>\n```\n\nThat\'s pretty much it!\n\n## Advanced usage\n\n### Email addresses should be unique\n\nWhen you\'re integrating with an OAuth provider,\nwe think that the user\'s email address is the best "primary key" when linking to your `User` model in your app.\nUnfortunately in Django, by default an email address is not required to be unique!\n**We strongly recommend you require email addresses to be unique in your app.**\n\n[As suggested by the Django docs](https://docs.djangoproject.com/en/4.0/topics/auth/customizing/#using-a-custom-user-model-when-starting-a-project),\none way to do this is to have your own `User` model:\n\n```python\n# In an app named "users", for example\nfrom django.contrib.auth.models import AbstractUser\n\nclass User(AbstractUser):\n    email = models.EmailField(unique=True)\n\n\n# In settings.py\nAUTH_USER_MODEL = \'users.User\'\n```\n\nYou\'ll also notice that there are no "email confirmation" or "email verification" flows in this library.\nThis is also intentional.\nYou can implement something like that yourself if you need to,\nbut the easier solution in our opinion is to use an OAuth provider you *trust to have done that already*.\nIf you look at our [provider examples](provider_examples) you\'ll notice how we often use provider APIs to get the email address which is "primary" and "verified" already.\nIf they\'ve already done that work,\nthen we can just use that information.\n\n### Handling OAuth errors\n\nThe most common error you\'ll run into is if an existing user clicks a login button,\nbut they haven\'t yet connected that provider to their account.\nFor security reasons,\nthe required flow here is that the user actually logs in with another method (however they signed up) and then *connects* the OAuth provider from a settings page.\n\nFor this error (and a couple others),\nthere is an error template that is rendered.\nYou can customize this by copying `oauth_login/error.html` to one of your own template directories:\n\n```html\n{% extends "base.html" %}\n\n{% block content %}\n<h1>OAuth Error</h1>\n<p>{{ oauth_error }}</p>\n{% endblock %}\n```\n\n### Connecting and disconnecting OAuth accounts\n\n### Using a saved access token\n\n```python\nimport requests\n\n# Get the OAuth connection for a user\nconnection = user.oauth_connections.get(provider_key="github")\n\n# If the token can expire, check and refresh it\nif connection.access_token_expired():\n    connection.refresh_access_token()\n\n# Use the token in an API call\ntoken = connection.oauth_token\nresponse = requests.get(...)\n```\n\n### Using the Django system check\n\nThis library comes with a Django system check to ensure you don\'t *remove* a provider from `settings.py` that is still in use in your database.\n\n## FAQs\n\n### How is this different from [other Django OAuth libraries](https://djangopackages.org/grids/g/oauth/)?\n\nThe short answer is that *it does less*.\n\nIn [django-allauth](https://github.com/pennersr/django-allauth)\n(maybe the most popular alternative)\nyou get all kinds of other features like managing multiple email addresses,\nemail verification,\na long list of supported providers,\nand a whole suite of forms/urls/views/templates/signals/tags.\nAnd in my experience,\nit\'s too much.\nIt often adds more complexity to your app than you actually need (or want) and honestly it can just be a lot to wrap your head around.\nPersonally, I don\'t like the way that your OAuth settings are stored in the database vs when you use `settings.py`,\nand the implications for doing it one way or another.\n\nThe other popular OAuth libraries have similar issues,\nand I think their *weight* outweighs their usefulness for 80% of the use cases.\n\n### Why aren\'t providers included in the library itself?\n\nOne thing you\'ll notice is that we don\'t have a long list of pre-configured providers in this library.\nInstead, we have some examples (which you can usually just copy, paste, and use) and otherwise encourage you to wire up the provider yourself.\nOften times all this means is finding the two OAuth URLs ("oauth/authorize" and "oauth/token") in their docs,\nand writing two class methods that do the actual work of getting the user\'s data (which is often customized anyway).\n\nWe\'ve written examples for the following providers:\n\n- [GitHub](provider_examples/github.py)\n- [GitLab](provider_examples/gitlab.py)\n- [Bitbucket](provider_examples/bitbucket.py)\n\nJust copy that code and paste it in your project.\nTweak as necessary!\n\nThis might sound strange at first.\nBut in the long run we think it\'s actually *much* more maintainable for both us (as library authors) and you (as app author).\nIf something breaks with a provider, you can fix it immediately!\nYou don\'t need to try to run changes through us or wait for an upstream update.\nYou\'re welcome to contribute an example to this repo,\nand there won\'t be an expectation that it "works perfectly for every use case until the end of time".\n',
    'author': 'Dave Gaeddert',
    'author_email': 'dave.gaeddert@dropseed.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dropseed/django-oauth-login',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



class OAuthError(Exception):
    """Base class for OAuth errors"""

    pass


class OAuthStateMismatchError(OAuthError):
    pass


class OAuthCannotDisconnectError(OAuthError):
    pass


class OAuthUserAlreadyExistsError(OAuthError):
    pass
